# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-05-05 21:26
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import orchestra.models.communication.mixins
import orchestra.utils.models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0060_auto_20160505_1349'),
    ]

    operations = [
        migrations.CreateModel(
            name='StaffBotRequest',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(
                    default=django.utils.timezone.now)),
                ('is_deleted', models.BooleanField(default=False)),
                ('required_role_counter', models.IntegerField()),
                ('request_cause', models.IntegerField(choices=[
                 (0, 'user'), (1, 'autostaff'), (2, 'restaff')])),
                ('project_description', models.TextField(blank=True, null=True)),
                ('status', models.IntegerField(choices=[
                 (0, 'processing'), (1, 'complete')], default=0)),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                           related_name='staffing_requests', to='orchestra.Task')),
            ],
            options={
                'abstract': False,
            },
            bases=(orchestra.models.communication.mixins.StaffBotRequestMixin,
                   orchestra.utils.models.DeleteMixin, models.Model),
        ),
        migrations.RemoveField(  # manually-reviewed
            model_name='staffingrequestinquiry',
            name='project_description',
        ),
        migrations.RemoveField(  # manually-reviewed
            model_name='staffingrequestinquiry',
            name='request_cause',
        ),
        migrations.RemoveField(  # manually-reviewed
            model_name='staffingrequestinquiry',
            name='required_role',
        ),
        migrations.RemoveField(  # manually-reviewed
            model_name='staffingrequestinquiry',
            name='status',
        ),
        migrations.RemoveField(  # manually-reviewed
            model_name='staffingrequestinquiry',
            name='task',
        ),
        migrations.AlterField(  # manually-reviewed
            model_name='staffingresponse',
            name='request',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to='orchestra.StaffingRequestInquiry'),
        ),
        migrations.AddField(
            model_name='staffingrequestinquiry',
            name='request',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE,
                                    related_name='inquiries', to='orchestra.StaffBotRequest'),
        ),
    ]



from enum import Enum

from django.db import transaction

from orchestra.core.errors import InvalidRevertError
from orchestra.models import Iteration
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.project_api.serializers import IterationSerializer
from orchestra.project_api.serializers import TaskSerializer
from orchestra.project_api.serializers import TaskAssignmentSerializer
from orchestra.utils.task_properties import assignment_history
from orchestra.utils.task_properties import get_iteration_history
from orchestra.utils.task_properties import get_latest_iteration


class RevertChange(Enum):
    UNCHANGED = 0
    REVERTED = 1
    DELETED = 2


@transaction.atomic
def revert_task_to_iteration(task_id, iteration_id,
                             revert_before=False, commit=False):
    """
    Reverts a task to or before a specified iteration.

    Args:
        task (orchestra.models.Task):
            The ID of the task to be reverted.
        iteration_id (datetime.datetime):
            The ID of the iteration before which to revert the task.
        revert_before (bool):
            If `revert_before` is indicated, the task will be reverted to a
            state before the specified iteration was created; otherwise, it
            will be reverted to a state in which the iteration is processing.
            This can only be true for the first iteration in an assignment.
        commit (bool):
            Determines whether the revert is actually carried out;
            otherwise, an audit trail is passively generated.

    Returns:
        audit (dict):
            An audit trail for the revert, e.g.,
            {
                'task': {...},
                'assignments': [
                    {
                        # worker_0 assignment
                        'assignment': {...},
                        'change': 1,
                        'iterations': [
                            {
                                'iteration': {...},
                                'change': 2
                            }
                        ]
                    },
                    {
                        # worker_1 assignment
                        'assignment': {...},
                        'change': 2,
                        'iterations': []
                    }
                ],
            }
    """
    task = Task.objects.get(id=task_id)
    revert_iteration = Iteration.objects.get(id=iteration_id)
    first_iteration = (
        revert_iteration.assignment.iterations
                        .order_by('start_datetime').first())
    if revert_before and revert_iteration != first_iteration:
        raise InvalidRevertError(
            "Cannot revert before an iteration that isn't the first of "
            "its assignment.")

    task_audit = _build_revert_audit(task, revert_iteration, revert_before)
    if commit:
        _revert_task_from_audit(task_audit)
    return task_audit


def _revert_task_from_audit(task_audit):
    task = Task.objects.get(id=task_audit['task']['id'])
    task.status = task_audit['reverted_status']
    task.save()
    for assignment_audit in task_audit['assignments']:
        _revert_assignment_from_audit(assignment_audit)


def _revert_assignment_from_audit(assignment_audit):
    assignment = TaskAssignment.objects.get(
        id=assignment_audit['assignment']['id'])
    if assignment_audit['change'] == RevertChange.DELETED.value:
        assignment.delete()
    elif assignment_audit['change'] == RevertChange.REVERTED.value:
        for iteration_audit in assignment_audit['iterations']:
            _revert_iteration_from_audit(iteration_audit)

        assignment.refresh_from_db()
        if get_latest_iteration(
                assignment).status == Iteration.Status.PROCESSING:
            assignment.status = TaskAssignment.Status.PROCESSING
        else:
            assignment.status = TaskAssignment.Status.SUBMITTED
        assignment.save()

        # TODO(jrbotros): Right now, we're leaving the latest assignment
        # `in_progress_task_data` alone since we would otherwise lose it;
        # when we implement archival rather than deletion, we should update
        # the latest assignment data to the submitted data of the iteration
        # reverted to.

    return assignment


def _revert_iteration_from_audit(iteration_audit):
    iteration = Iteration.objects.get(
        id=iteration_audit['iteration']['id'])
    if iteration_audit['change'] == RevertChange.DELETED.value:
        iteration.delete()
    elif iteration_audit['change'] == RevertChange.REVERTED.value:
        iteration = Iteration.objects.get(
            id=iteration_audit['iteration']['id'])
        iteration.status = Iteration.Status.PROCESSING
        iteration.end_datetime = None
        iteration.submitted_data = {}
        iteration.save()


def _build_revert_audit(task, revert_iteration, revert_before):
    if task.status == Task.Status.ABORTED:
        raise InvalidRevertError('Cannot revert aborted task.')

    task_audit = {
        'task': TaskSerializer(task).data,
        'assignments': [],
    }

    for assignment in assignment_history(task).all():
        assignment_audit = (
            _build_assignment_revert_audit(
                assignment, revert_iteration, revert_before))
        task_audit['assignments'].append(assignment_audit)

    task_audit['reverted_status'] = _reverted_task_status(
        task_audit, revert_before)
    return task_audit


def _build_assignment_revert_audit(assignment, revert_iteration,
                                   revert_before):
    assignment_audit = {
        'assignment': TaskAssignmentSerializer(assignment).data,
        'change': RevertChange.UNCHANGED.value,
        'iterations': []
    }

    for iteration in assignment.iterations.order_by('start_datetime').all():
        iteration_audit = (
            _build_iteration_revert_audit(
                iteration, revert_iteration, revert_before))
        assignment_audit['iterations'].append(iteration_audit)

    changed_iterations = _parse_changed_items(
        assignment_audit['iterations'], 'iteration')
    deleted_iterations = changed_iterations[RevertChange.DELETED.value]
    reverted_iterations = changed_iterations[RevertChange.REVERTED.value]

    if len(deleted_iterations) == assignment.iterations.count():
        assignment_audit['change'] = RevertChange.DELETED.value
    elif deleted_iterations or reverted_iterations:
        assignment_audit['change'] = RevertChange.REVERTED.value
    return assignment_audit


def _build_iteration_revert_audit(iteration, revert_iteration, revert_before):
    iteration_audit = {
        'iteration': IterationSerializer(iteration).data,
        'change': RevertChange.UNCHANGED.value,
    }
    if revert_before:
        if (iteration.id == revert_iteration.id or
                iteration.start_datetime >= revert_iteration.start_datetime):
            iteration_audit['change'] = RevertChange.DELETED.value
    else:
        if iteration.id == revert_iteration.id:
            iteration_audit['change'] = RevertChange.REVERTED.value
        elif iteration.start_datetime > revert_iteration.start_datetime:
            iteration_audit['change'] = RevertChange.DELETED.value
    return iteration_audit


def _reverted_task_status(task_audit, revert_before):
    """
    Reverts the status of an otherwise-reverted task.

    Args:
        task (dict):
            Audit containing task data to be changed upon revert.

    Returns:
        status (orchestra.models.Task.Status):
            The status of the task if it were reverted.
    """
    task = Task.objects.get(id=task_audit['task']['id'])

    flattened_iterations = [
        iteration_audit
        for assignment_audit in task_audit['assignments']
        for iteration_audit in assignment_audit['iterations']]
    changed_items = _parse_changed_items(flattened_iterations, 'iteration')

    latest_iterations = (
        get_iteration_history(task, reverse=True)
        .exclude(id__in=changed_items[RevertChange.DELETED.value]))

    num_iterations = latest_iterations.count()
    if num_iterations == 0:
        return Task.Status.AWAITING_PROCESSING
    elif revert_before:
        # Reverting before the first iteration in an assignment means the task
        # is pending review, since at least one iteration exists
        return Task.Status.PENDING_REVIEW
    else:
        # Revert to a processing iteration state
        if num_iterations == 1:
            return Task.Status.PROCESSING
        else:
            previous_status = latest_iterations[1].status
            if previous_status == Iteration.Status.REQUESTED_REVIEW:
                return Task.Status.REVIEWING
            else:
                return Task.Status.POST_REVIEW_PROCESSING


def _parse_changed_items(items, item_key):
    changed_items = {change.value: [] for change in RevertChange}
    for item in items:
        changed_items[item['change']].append(item[item_key]['id'])
    return changed_items



import json

from django.conf import settings
from django.contrib.auth.models import Group
from django.core.urlresolvers import reverse
from django.test import Client

from orchestra.core.errors import InvalidRevertError
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import Project
from orchestra.project_api.serializers import IterationSerializer
from orchestra.project_api.serializers import TaskSerializer
from orchestra.project_api.serializers import TaskAssignmentSerializer
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import setup_complete_task
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.tests.helpers.fixtures import WorkerFactory
from orchestra.tests.helpers.iterations import verify_iterations
from orchestra.utils.load_json import load_encoded_json
from orchestra.utils.revert import RevertChange
from orchestra.utils.task_lifecycle import create_subsequent_tasks
from orchestra.utils.task_properties import assignment_history
from orchestra.utils.task_properties import current_assignment
from orchestra.utils.task_properties import get_iteration_history
from orchestra.utils.task_properties import get_latest_iteration


class ProjectManagementAPITestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.project_admin_group = Group.objects.create(name='project_admins')
        user = (UserFactory(username='project_management',
                            first_name='project_management',
                            last_name='project_management',
                            password='project_management',
                            email='project_management@b12.io'))
        self.worker = WorkerFactory(user=user)
        user.groups.add(self.project_admin_group)

        self.api_client = Client()
        self.api_client.login(
            username='project_management', password='project_management')

    def test_project_information_api(self):
        project = self.projects['project_management_project']
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:'
                'project_information'),
            json.dumps({
                'project_id': project.id,
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)

        # Skipping the `tasks` key/value pair for this test
        expected_project = {
            'task_class': project.task_class,
            'start_datetime': '2015-10-12T00:00:00Z',
            'team_messages_url': None,
            'admin_url': settings.ORCHESTRA_URL + reverse(
                'admin:orchestra_project_change',
                args=(project.id,)),
            'priority': project.priority,
            'project_data': project.project_data,
            'short_description': project.short_description,
            'id': project.id,
            'workflow_slug': project.workflow_version.workflow.slug
        }
        self.assertEqual(
            expected_project,
            {k: returned['project'][k] for k in expected_project.keys()})

        sample_task = returned['tasks']['step1']
        task = project.tasks.get(step__slug='step1')
        expected_task = {
            'status': 'Post-review Processing',
            'start_datetime': '2015-10-12T01:00:00Z',
            'admin_url': settings.ORCHESTRA_URL + reverse(
                'admin:orchestra_task_change',
                args=(task.id,)),
            'latest_data': {
                'test_key': 'test_value'
            },
            'project': task.project.id,
            'step_slug': 'step1',
            'id': task.id
        }

        self.assertEqual(
            expected_task,
            {k: sample_task[k] for k in expected_task.keys()})

        sample_assignment = sample_task['assignments'][0]
        assignment = assignment_history(task)[0]
        expected_assignment = {
            'status': 'Submitted',
            'start_datetime': '2015-10-12T02:00:00Z',
            'task': assignment.task.id,
            'admin_url': settings.ORCHESTRA_URL + reverse(
                'admin:orchestra_taskassignment_change',
                args=(assignment.id,)),
            'worker': {
                'username': assignment.worker.user.username,
                'first_name': assignment.worker.user.first_name,
                'last_name': assignment.worker.user.last_name,
                'id': assignment.worker.id,
            },
            'iterations': [
                {
                    'id': assignment.iterations.first().id,
                    'admin_url': settings.ORCHESTRA_URL + reverse(
                        'admin:orchestra_iteration_change',
                        args=(assignment.iterations.first().id,)),
                    'assignment': assignment.id,
                    'start_datetime': '2015-10-12T02:00:00Z',
                    'end_datetime': '2015-10-12T03:00:00Z',
                    'status': 'Requested Review',
                    'submitted_data': {}
                }
            ],
            'in_progress_task_data': {},
            'id': assignment.id
        }

        self.assertEqual(
            expected_assignment,
            {k: sample_assignment[k] for k in expected_assignment.keys()})

    def test_reassign_assignment_api(self):
        task = self.tasks['project_management_task']
        # Reassign entry-level assignment to entry-level worker
        entry_assignment = task.assignments.get(assignment_counter=0)
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:reassign_assignment'),
            json.dumps({
                'worker_username': self.workers[0].user.username,
                'assignment_id': entry_assignment.id
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        entry_assignment.refresh_from_db()
        self.assertEqual(entry_assignment.worker, self.workers[0])

        # Reassign entry-level assignment to reviewer
        entry_assignment.refresh_from_db()
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:reassign_assignment'),
            json.dumps({
                'worker_username': self.workers[1].user.username,
                'assignment_id': entry_assignment.id
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        entry_assignment.refresh_from_db()
        self.assertEqual(entry_assignment.worker, self.workers[1])

        # Attempt to reassign assignment to worker already working on task
        first_review_assignment = task.assignments.get(assignment_counter=1)
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:reassign_assignment'),
            json.dumps({
                'worker_username': self.workers[1].user.username,
                'assignment_id': first_review_assignment.id
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        first_review_assignment.refresh_from_db()
        self.assertTrue(first_review_assignment.worker, self.workers[1])
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Worker already assigned to this task.')

        # Attempt to reassign review assignment to entry-level worker
        first_review_assignment = task.assignments.get(assignment_counter=1)
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:reassign_assignment'),
            json.dumps({
                'worker_username': self.workers[0].user.username,
                'assignment_id': first_review_assignment.id
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        first_review_assignment.refresh_from_db()
        self.assertTrue(first_review_assignment.worker, self.workers[0])
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Worker not certified for this assignment.')

        # Reassign review assignment to another reviewer
        second_review_assignment = task.assignments.get(assignment_counter=2)
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:reassign_assignment'),
            json.dumps({
                'worker_username': self.workers[3].user.username,
                'assignment_id': second_review_assignment.id
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        second_review_assignment.refresh_from_db()
        self.assertEqual(second_review_assignment.worker, self.workers[3])

    def test_complete_and_skip_task_api(self):
        task = self.tasks['project_management_task']
        response = self.api_client.post(
            reverse(
                'orchestra:orchestra:project_management:'
                'complete_and_skip_task'),
            json.dumps({
                'task_id': task.id,
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        task.refresh_from_db()
        self.assertEqual(task.status, Task.Status.COMPLETE)
        for assignment in task.assignments.all():
            self.assertEqual(
                assignment.status, TaskAssignment.Status.SUBMITTED)

        # Check that dependent tasks have already been created
        # TODO(jrbotros): Create a `get_dependent_tasks` function
        num_tasks = task.project.tasks.count()
        create_subsequent_tasks(task.project)
        self.assertEqual(task.project.tasks.count(), num_tasks)

    def test_end_project_api(self):
        project = self.projects['project_to_end']
        response = self.api_client.post(
            reverse('orchestra:orchestra:project_management:end_project'),
            json.dumps({
                'project_id': project.id,
            }),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)
        project.refresh_from_db()
        self.assertEqual(project.status, Project.Status.ABORTED)
        for task in Task.objects.filter(project=project):
            self.assertEqual(task.status, Task.Status.ABORTED)

    def test_invalid_revert_aborted_task(self):
        task = setup_complete_task(self)
        task.status = Task.Status.ABORTED
        task.save()

        with self.assertRaises(InvalidRevertError):
            self._revert_task(
                task, task.assignments.first().iterations.first(),
                revert_before=False, commit=True)

        task.delete()

    def test_invalid_revert_before(self):
        task = setup_complete_task(self)
        task.status = Task.Status.ABORTED
        task.save()

        # Ensure reviewer assignment has more than one iteration
        reviewer_assignment = assignment_history(task).last()
        self.assertGreater(reviewer_assignment.iterations.count(), 1)

        # Attempt to revert before an iteration that isn't the assignment's
        # first
        with self.assertRaises(InvalidRevertError):
            self._revert_task(
                task, get_latest_iteration(reviewer_assignment),
                revert_before=True, commit=True)

        task.delete()

    def test_revert_second_review(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.REVIEWING

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.UNCHANGED.value, RevertChange.REVERTED.value),
            iteration_changes=(
                (RevertChange.UNCHANGED.value, RevertChange.UNCHANGED.value),
                (RevertChange.UNCHANGED.value, RevertChange.REVERTED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[3],
            num_iterations=4,
            task_status=reverted_status,
            latest_data={'test': 'reviewer_accept'},
            expected_audit=expected_audit)

        task.delete()

    def test_revert_post_review_processing(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.POST_REVIEW_PROCESSING

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.REVERTED.value, RevertChange.REVERTED.value),
            iteration_changes=(
                (RevertChange.UNCHANGED.value, RevertChange.REVERTED.value),
                (RevertChange.UNCHANGED.value, RevertChange.DELETED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[2],
            num_iterations=3,
            task_status=reverted_status,
            latest_data={'test': 'entry_resubmit'},
            expected_audit=expected_audit)

        task.delete()

    def test_revert_first_review(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.REVIEWING

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.REVERTED.value, RevertChange.REVERTED.value),
            iteration_changes=(
                (RevertChange.UNCHANGED.value, RevertChange.DELETED.value),
                (RevertChange.REVERTED.value, RevertChange.DELETED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[1],
            num_iterations=2,
            task_status=reverted_status,
            latest_data={'test': 'reviewer_accept'},
            expected_audit=expected_audit)

        task.delete()

    def test_revert_pending_review(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.PENDING_REVIEW

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.REVERTED.value, RevertChange.DELETED.value),
            iteration_changes=(
                (RevertChange.UNCHANGED.value, RevertChange.DELETED.value),
                (RevertChange.DELETED.value, RevertChange.DELETED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[1],
            num_iterations=1,
            task_status=reverted_status,
            latest_data={'test': 'entry_resubmit'},
            expected_audit=expected_audit,
            revert_before=True)

        task.delete()

    def test_revert_processing(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.PROCESSING

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.REVERTED.value, RevertChange.DELETED.value),
            iteration_changes=(
                (RevertChange.REVERTED.value, RevertChange.DELETED.value),
                (RevertChange.DELETED.value, RevertChange.DELETED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[0],
            num_iterations=1,
            task_status=reverted_status,
            latest_data={'test': 'entry_resubmit'},
            expected_audit=expected_audit)

        task.delete()

    def test_revert_awaiting_processing(self):
        task = setup_complete_task(self)
        reverted_status = Task.Status.AWAITING_PROCESSING

        expected_audit = self._expected_audit(
            task,
            reverted_status=reverted_status,
            assignment_changes=(
                RevertChange.DELETED.value, RevertChange.DELETED.value),
            iteration_changes=(
                (RevertChange.DELETED.value, RevertChange.DELETED.value),
                (RevertChange.DELETED.value, RevertChange.DELETED.value)
            ))

        self._test_reverted_task(
            task,
            iteration=get_iteration_history(task).all()[0],
            num_iterations=0,
            task_status=reverted_status,
            latest_data={'test': 'entry_resubmit'},
            expected_audit=expected_audit,
            revert_before=True)

        task.delete()

    def _revert_task(self, task, iteration, revert_before, commit):
        return self.api_client.post(
            reverse('orchestra:orchestra:project_management:revert_task'),
            json.dumps({
                'task_id': task.id,
                # Convert datetime to timestamp
                'iteration_id': iteration.id,
                'revert_before': revert_before,
                'commit': commit,
            }),
            content_type='application/json')

    def _test_reverted_task(self, task, iteration, num_iterations,
                            task_status, latest_data, expected_audit,
                            revert_before=False):
        response = self._revert_task(
            task, iteration, revert_before=revert_before, commit=False)
        self.assertEqual(response.status_code, 200)
        task.refresh_from_db()
        fake_audit = load_encoded_json(response.content)
        self.assertEqual(fake_audit, expected_audit)

        self.assertEqual(task.status, Task.Status.COMPLETE)
        self.assertEqual(task.assignments.count(), 2)
        for assignment in task.assignments.all():
            self.assertEqual(assignment.iterations.count(), 2)

        response = self._revert_task(
            task, iteration, revert_before=revert_before, commit=True)
        self.assertEqual(response.status_code, 200)
        task.refresh_from_db()
        audit = load_encoded_json(response.content)
        self.assertEqual(audit, fake_audit)
        task.refresh_from_db()
        self.assertEqual(task.status, task_status)
        self.assertEqual(
            get_iteration_history(task).count(), num_iterations)

        verify_iterations(task.id)

        if num_iterations:
            self.assertEqual(
                current_assignment(task).in_progress_task_data,
                latest_data)

        verify_iterations(task.id)

    def _expected_audit(self, complete_task, reverted_status=None,
                        assignment_changes=None, iteration_changes=None):
        assignments = assignment_history(complete_task)
        iterations = get_iteration_history(complete_task)
        audit = {
            'task': TaskSerializer(complete_task).data,
            'assignments': [
                {
                    'assignment': (
                        TaskAssignmentSerializer(assignments.first()).data),
                    'change': 0,
                    'iterations': [
                        {
                            'change': 0,
                            'iteration': (
                                IterationSerializer(iterations.all()[0]).data)
                        },
                        {
                            'change': 0,
                            'iteration': (
                                IterationSerializer(iterations.all()[2]).data)
                        }
                    ]
                },
                {
                    'assignment': (
                        TaskAssignmentSerializer(assignments.last()).data),
                    'change': 0,
                    'iterations': [
                        {
                            'change': 0,
                            'iteration': (
                                IterationSerializer(iterations.all()[1]).data)
                        },
                        {
                            'change': 0,
                            'iteration': (
                                IterationSerializer(iterations.all()[3]).data)
                        }
                    ]
                }
            ],
        }
        if reverted_status is not None:
            audit['reverted_status'] = reverted_status
        if assignment_changes is not None:
            for i, assignment_change in enumerate(assignment_changes):
                audit['assignments'][i]['change'] = assignment_change
        if iteration_changes is not None:
            for i, changes_per_assignment in enumerate(iteration_changes):
                for j, iteration_change in enumerate(changes_per_assignment):
                    audit['assignments'][i][
                        'iterations'][j]['change'] = iteration_change
        return audit



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0015_auto_20150807_1929'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='slack_group_id',
            field=models.CharField(null=True, blank=True, max_length=200),
        ),
        migrations.AddField(
            model_name='worker',
            name='slack_username',
            field=models.CharField(null=True, blank=True, max_length=200),
        ),
        migrations.AlterField(
            model_name='project',
            name='workflow_slug',
            field=models.CharField(max_length=200, choices=[('website_enhancement', 'Website enhancement'), (
                'website_design_v2', 'Website design v2'), ('website_design', 'Website design')]),
        ),
        migrations.AlterField(
            model_name='task',
            name='step_slug',
            field=models.CharField(max_length=200, choices=[('export_enhancement_step', 'Export enhancement'), ('website_enhancement', 'Enhance'), ('design_v2', 'Design'), ('client_interview_v2', 'Client Interview'), ('seo_v2', 'SEO Optimization'), (
                'communication_delivery_v2', 'Communication & Delivery'), ('export_v2', 'Export design'), ('media_extraction_v2', 'Media Extraction'), ('ramp_down_v2', 'Ramping Down'), ('ramp_up_v2', 'Ramping Up'), ('design', 'Design'), ('media_extraction', 'Media Extraction'), ('export', 'Export design')]),
        ),
    ]



import json

from beanstalk_dispatch import ARGS
from beanstalk_dispatch import FUNCTION
from beanstalk_dispatch import KWARGS


def create_request_body(function_name, *args, **kwargs):
    return json.dumps({FUNCTION: function_name, ARGS: args, KWARGS: kwargs})



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-27 14:57
from __future__ import unicode_literals

from django.db import migrations
import phonenumber_field.modelfields


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0046_auto_20160426_2034'),
    ]

    operations = [
        migrations.AddField(
            model_name='worker',
            name='phone',
            field=phonenumber_field.modelfields.PhoneNumberField(
                max_length=128, null=True),
        ),
    ]



from django.dispatch import receiver
from django.dispatch import Signal

from orchestra.models import Worker
from orchestra.models import CommunicationPreference

# A user has activated his or her account.
orchestra_user_registered = Signal(providing_args=['user', 'request'])


@receiver(orchestra_user_registered)
def add_worker_for_new_users(sender, user, request, **kwargs):
    worker, _ = Worker.objects.get_or_create(user=user)
    CommunicationPreference.objects.get_or_create_all_types(worker)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-08-02 03:01
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0064_workercertification_staffbot_enabled'),
    ]

    operations = [
        migrations.AlterField(
            model_name='staffingresponse',
            name='request_inquiry',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                    related_name='responses', to='orchestra.StaffingRequestInquiry'),
        ),
    ]



from django.core.urlresolvers import reverse

from orchestra.tests.helpers import OrchestraAuthenticatedTestCase
from orchestra.tests.helpers.fixtures import StaffingResponseFactory
from orchestra.tests.helpers.fixtures import WorkerFactory


class StaffRequestBase(OrchestraAuthenticatedTestCase):
    template = ''
    url_reverse = ''
    __test__ = False

    def setUp(self):
        super().setUp()
        self.user = self.authenticate_user()
        self.staffing_response = StaffingResponseFactory(
            request_inquiry__request__task__step__is_human=True,
            request_inquiry__communication_preference__worker__user=self.user)
        self.url_kwargs = {
            'staffing_request_inquiry_id': (self.staffing_response.
                                            request_inquiry.pk)
        }
        self.url = reverse(self.url_reverse, kwargs=self.url_kwargs)

    def assert_response(self, response, status_code=200, template=None):
        self.assertEqual(response.status_code, status_code)
        if template is None:
            template = self.template
        self.assertTemplateUsed(response, template)

    def test_get(self):
        response = self.request_client.get(self.url)
        self.assert_response(response)

    def test_unauthenticated_request(self):
        new_worker = WorkerFactory()
        comm_pref = (self.staffing_response.
                     request_inquiry.communication_preference)
        comm_pref.worker = new_worker
        comm_pref.save()

        response = self.request_client.get(self.url)
        self.assert_response(response, status_code=404,
                             template='orchestra/error.html')

    def test_unauthorized_request(self):
        self.request_client.logout()

        response = self.request_client.get(self.url)
        login_url = '{}?next={}'.format(reverse('auth_login'), self.url)
        self.assertRedirects(response, login_url)


class AcceptStaffRequestTest(StaffRequestBase):
    __test__ = True
    template = 'communication/staffing_request_accepted.html'
    url_reverse = 'orchestra:communication:accept_staffing_request_inquiry'

    def test_reject_after_accept(self):
        response = self.request_client.get(self.url)
        self.assert_response(response)

        reject_url = reverse(
            'orchestra:communication:reject_staffing_request_inquiry',
            kwargs=self.url_kwargs)

        response = self.request_client.get(reject_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response,
            'communication/staffing_response_not_permitted.html')


class RejectStaffRequestTest(StaffRequestBase):
    __test__ = True
    template = 'communication/staffing_request_rejected.html'
    url_reverse = 'orchestra:communication:reject_staffing_request_inquiry'



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-29 11:27
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0056_create_new_task_communication_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='staffingrequest',
            name='communication_preference',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to='orchestra.CommunicationPreference'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0013_auto_20150707_0054'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='status',
            field=models.IntegerField(
                default=0, choices=[(-1, 'Aborted'), (0, 'Active')]),
        ),
        migrations.AlterField(
            model_name='task',
            name='step_slug',
            field=models.CharField(choices=[('design', 'Design'), ('media_extraction', 'Media Extraction'), ('export', 'Export design'), (
                'export_enhancement_step', 'Export enhancement'), ('website_enhancement', 'Enhance')], max_length=200),
        ),
    ]



import random
import string

from django.conf import settings
from django.utils.text import slugify
from slacker import BaseAPI
from slacker import Error as SlackError
from slacker import Slacker

from orchestra.utils.decorators import run_if
from orchestra.communication.errors import SlackFormatError

import logging
logger = logging.getLogger(__name__)

# Types of responses we can send to slack
VALID_RESPONSE_TYPES = {'ephemeral', 'in_channel'}
_request = BaseAPI._request


def _silent_request(*args, **kwargs):
    """
    Attempt to make Slack API request and ignore if an exception is thrown.

    TODO(jrbotros): this silences all errors, but we likely will want to be
    able to surface errors in some cases in the future
    """
    try:
        return _request(*args, **kwargs)
    except SlackError:
        logger.exception('Slack API Error')


BaseAPI._request = _silent_request


class OrchestraSlackService(object):
    """
    Wrapper slack service to allow easy swapping and mocking out of API.
    """

    def __init__(self, api_key=None):
        if not api_key:
            api_key = settings.SLACK_EXPERTS_API_KEY
        self._service = Slacker(api_key)
        for attr_name in ('chat', 'groups', 'users'):
            setattr(self, attr_name, getattr(self._service, attr_name))


@run_if('ORCHESTRA_SLACK_EXPERTS_ENABLED')
def get_slack_user_id(slack_username):
    slack = OrchestraSlackService()
    slack_user_id = slack.users.get_user_id(slack_username)
    return slack_user_id


@run_if('ORCHESTRA_SLACK_EXPERTS_ENABLED')
def add_worker_to_project_team(worker, project):
    slack = OrchestraSlackService()
    try:
        user_id = slack.users.get_user_id(worker.slack_username)
        response = slack.groups.invite(project.slack_group_id, user_id)
        if not response.body.get('already_in_group'):
            welcome_message = (
                '<@{}|{}> has been added to the team. '
                'Welcome aboard!').format(user_id, worker.slack_username)
            slack.chat.post_message(project.slack_group_id, welcome_message)
    except:
        # TODO(jrbotros): for now, using slack on a per-worker basis is
        # optional; we'll want to rethink this in the future
        pass


@run_if('ORCHESTRA_SLACK_EXPERTS_ENABLED')
def create_project_slack_group(project):
    """
    Create slack channel for project team communication
    """
    slack = OrchestraSlackService()
    response = slack.groups.create(_project_slack_group_name(project))
    project.slack_group_id = response.body['group']['id']
    slack.groups.set_topic(project.slack_group_id, project.short_description)
    slack.groups.set_purpose(project.slack_group_id,
                             'Discussing work on `{}`'.format(
                                 project.short_description))
    project.save()
    return project.slack_group_id


def _random_string():
    return ''.join(
        random.choice(string.ascii_lowercase + string.digits)
        for idx in range(4))


def _project_slack_group_name(project):
    """
    Return a unique and readable identifier for project slack groups.

    Slack group names are capped at 21 characters in length.
    """
    name = None
    # The human-readable portion of the name (16 characters) involves
    # slugifying the project short description.
    descriptor = slugify(project.short_description)[:16].strip('-')
    slack = OrchestraSlackService()
    groups = {group['name'] for group in slack.groups.list().body['groups']}
    while True:
        # Add 4 characters of randomness (~1.68 million permutations).
        name = '{}-{}'.format(descriptor, _random_string())
        if name not in groups:
            break
    return name


def format_slack_message(text,
                         attachments=None,
                         response_type='ephemeral'):
    """
    Args:
        text (str):
            Plain text message to send
        attachments (dict):
           See https://api.slack.com/docs/attachments
        response_type (string):
            Should be `in_channel` or `ephemeral`. See
            https://api.slack.com/slash-commands
    Returns:
        formatted_message (dict):
            A formatted message to send via the slack client
    """
    if response_type not in VALID_RESPONSE_TYPES:
        raise SlackFormatError(
            'Response type {} is invalid'.format(response_type))
    return {
        'response_type': response_type,
        'text': text,
        'attachments': attachments,
    }



import boto
import json

from beanstalk_dispatch import ARGS
from beanstalk_dispatch import FUNCTION
from beanstalk_dispatch import KWARGS
from beanstalk_dispatch.client import schedule_function
from django.test import TestCase
from django.test import override_settings
from moto import mock_sqs

CALL_COUNTER = 0


def counter_incrementer(first_arg, second_arg=None):
    global CALL_COUNTER
    CALL_COUNTER += first_arg
    if second_arg:
        CALL_COUNTER += second_arg


DISPATCH_SETTINGS = {
    'BEANSTALK_DISPATCH_TABLE': {
        'the_counter': ('beanstalk_dispatch.tests.'
                        'test_client.counter_incrementer')
    }
}


@mock_sqs
@override_settings(
    BEANSTALK_DISPATCH_SQS_KEY='', BEANSTALK_DISPATCH_SQS_SECRET='')
class ClientTestCase(TestCase):

    def setUp(self):
        self.queue_name = 'testing-queue'

    def test_async_function_scheduling(self):
        # Check the message on the queue.
        sqs_connection = boto.connect_sqs('', '')
        sqs_connection.create_queue(self.queue_name)
        queue = sqs_connection.get_queue(self.queue_name)
        messages = queue.get_messages()
        self.assertEqual(len(messages), 0)

        # Schedule a function.
        schedule_function(
            self.queue_name, 'a-function', '1', '2', kwarg1=1, kwarg2=2)

        messages = queue.get_messages()
        self.assertEqual(len(messages), 1)

        # For some reason, boto base64-encodes the messages, but moto does
        # not.  Life.
        self.assertEqual(
            json.loads(messages[0].get_body()),
            {FUNCTION: 'a-function', ARGS: ['1', '2'], KWARGS: {
                'kwarg1': 1, 'kwarg2': 2}})

    @override_settings(
        BEANSTALK_DISPATCH_EXECUTE_SYNCHRONOUSLY=True,
        **DISPATCH_SETTINGS
    )
    def test_sync_function_scheduling(self):
        # Schedule a function.
        schedule_function(self.queue_name, 'the_counter', 1, second_arg=5)

        self.assertEqual(CALL_COUNTER, 6)



from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.permissions import BasePermission
from rest_framework_httpsignature.authentication import SignatureAuthentication

import logging
logger = logging.getLogger(__name__)


class SignedUser(AnonymousUser):
    pass


class IsSignedUser(BasePermission):

    def has_permission(self, request, view):
        return request.user and isinstance(request.user, SignedUser)


class OrchestraProjectAPIAuthentication(SignatureAuthentication):
    API_KEY_HEADER = 'X-Api-Key'

    def fetch_user_data(self, api_key):
        try:
            credentials = settings.ORCHESTRA_PROJECT_API_CREDENTIALS
            return (SignedUser(), credentials[api_key])
        except KeyError:
            raise AuthenticationFailed('Unable to find API key')
        except AttributeError:
            raise AuthenticationFailed(
                'Please configure settings.ORCHESTRA_PROJECT_API_CREDENTIALS')



from django.contrib.auth.models import User
from orchestra.models import Certification
from orchestra.models import Worker
from orchestra.models import WorkerCertification


def load(workflow_version):
    """ Loads workers for running the journalism workflow. """
    for user_data in USERS:
        # Create the user/worker objects and set the password.
        username = user_data.pop('username')
        password = user_data.pop('password')
        certifications = user_data.pop('certifications')
        user, _ = User.objects.update_or_create(
            username=username,
            defaults=user_data,
        )
        user.set_password(password)
        user.save()
        worker, _ = Worker.objects.update_or_create(user=user)

        # Grant the worker the desired certifications.
        for certification_slug, certification_role in certifications:
            certification = Certification.objects.get(
                slug=certification_slug)
            WorkerCertification.objects.update_or_create(
                certification=certification,
                worker=worker,
                task_class=WorkerCertification.TaskClass.REAL,
                role=certification_role)


USERS = [
    {
        'username': 'journalism-editor',
        'first_name': 'Journalism',
        'last_name': 'Editor',
        'is_active': True,
        'is_superuser': False,
        'is_staff': False,
        'email': 'noreply@example.org',
        'password': 'editor',
        'certifications': [
            ('editor', WorkerCertification.Role.ENTRY_LEVEL),
        ]
    },
    {
        'username': 'journalism-reporter-1',
        'first_name': 'Journalism',
        'last_name': 'Reporter1',
        'is_active': True,
        'is_superuser': False,
        'is_staff': False,
        'password': 'reporter',
        'email': 'noreply@example.org',
        'certifications': [
            ('reporter', WorkerCertification.Role.ENTRY_LEVEL),
        ]
    },
    {
        'username': 'journalism-reporter-2',
        'first_name': 'Journalism',
        'last_name': 'Reporter2',
        'is_active': True,
        'is_superuser': False,
        'is_staff': False,
        'password': 'reporter',
        'email': 'noreply@example.org',
        'certifications': [
            ('reporter', WorkerCertification.Role.ENTRY_LEVEL),
            ('reporter', WorkerCertification.Role.REVIEWER),
        ],
    },
    {
        'username': 'journalism-photographer-1',
        'first_name': 'Journalism',
        'last_name': 'Photographer1',
        'is_active': True,
        'is_superuser': False,
        'is_staff': False,
        'password': 'photographer',
        'email': 'noreply@example.org',
        'certifications': [
            ('photographer', WorkerCertification.Role.ENTRY_LEVEL),
        ]
    },
    {
        'username': 'journalism-photographer-2',
        'first_name': 'Journalism',
        'last_name': 'Photographer2',
        'is_active': True,
        'is_superuser': False,
        'is_staff': False,
        'password': 'photographer',
        'email': 'noreply@example.org',
        'certifications': [
            ('photographer', WorkerCertification.Role.ENTRY_LEVEL),
            ('photographer', WorkerCertification.Role.REVIEWER),
        ]
    },
    {
        'username': 'journalism-copy-editor',
        'first_name': 'Journalism',
        'last_name': 'CopyEditor',
        'is_active': True,
        'is_superuser': False,
        'is_staff': False,
        'password': 'copy-editor',
        'email': 'noreply@example.org',
        'certifications': [
            ('copy_editor', WorkerCertification.Role.ENTRY_LEVEL),
        ]
    },
]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-19 22:28
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0042_mark_deleted_timeentries_no_worker'),
    ]

    operations = [
        migrations.AddField(
            model_name='tasktimer',
            name='description',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
    ]



import json

from base64 import b64encode
from beanstalk_dispatch import ARGS
from beanstalk_dispatch import FUNCTION
from beanstalk_dispatch import KWARGS
from beanstalk_dispatch.common import create_request_body
from beanstalk_dispatch.safe_task import SafeTask
from django.core.urlresolvers import reverse
from django.test import Client
from django.test import TestCase
from django.test import override_settings

from orchestra.utils.load_json import load_encoded_json

CALL_COUNTER = 0


def counter_incrementer(first_arg, second_arg=None):
    global CALL_COUNTER
    CALL_COUNTER += first_arg
    if second_arg:
        CALL_COUNTER += second_arg


class CounterIncrementerTask(SafeTask):

    def run(self, *args, **kwargs):
        counter_incrementer(*args, **kwargs)


class BadTaskClass(object):
    pass


DISPATCH_SETTINGS = {
    'BEANSTALK_DISPATCH_TABLE': {
        'the_counter': ('beanstalk_dispatch.tests.'
                        'test_dispatcher.counter_incrementer'),
        'the_counter_task': ('beanstalk_dispatch.tests.'
                             'test_dispatcher.CounterIncrementerTask'),
        'bad_task_class': ('beanstalk_dispatch.tests.'
                           'test_dispatcher.BadTaskClass'),
        'bad_function_pointer': 'nothing-to-see-here',
    }
}


@override_settings(**DISPATCH_SETTINGS)
class DispatcherTestCase(TestCase):

    """ Test the server-side function dispatcher.

    In these tests, we base64-encode every message we send to the server
    because this is what boto does.
    """

    def setUp(self):
        global CALL_COUNTER
        CALL_COUNTER = 0
        self.client = Client()
        self.url = reverse('beanstalk_dispatcher')

    def test_no_get(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 405)

    @override_settings(BEANSTALK_DISPATCH_TABLE=None)
    def test_no_dispatch(self):
        response = self.client.post(
            self.url, b64encode(
                create_request_body('some_func').encode('ascii')),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(load_encoded_json(response.content),
                         {'message': 'No beanstalk dispatch table configured',
                          'error': 400})

    def test_missing_function(self):
        response = self.client.post(
            self.url,
            b64encode(create_request_body('nonexistent_func').encode('ascii')),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            load_encoded_json(response.content),
            {
                'message': 'Requested function not found: nonexistent_func',
                'error': 400
            })

    def test_invalid_task_class(self):
        response = self.client.post(
            self.url,
            b64encode(create_request_body(
                'bad_task_class', 'test-queue', {}).encode('ascii')),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            load_encoded_json(response.content),
            {
                'message': ('Requested task is not a SafeTask'
                            ' subclass: bad_task_class'),
                'error': 400
            })

    def test_invalid_function_pointer(self):
        response = self.client.post(
            self.url,
            b64encode(create_request_body(
                'bad_function_pointer', 'test-queue', {}).encode('ascii')),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            load_encoded_json(response.content),
            {
                'message': 'Unable to locate function: nothing-to-see-here',
                'error': 400
            })

    def test_malformed_request(self):
        keys = {FUNCTION, ARGS, KWARGS}
        for missing_key in keys:
            request_body = {key: 'test' for key in
                            keys - {missing_key}}
            response = self.client.post(
                self.url,
                b64encode(json.dumps(request_body).encode('ascii')),
                content_type='application/json')
            self.assertEqual(response.status_code, 400)
            self.assertEqual(load_encoded_json(response.content), {
                'message': 'Please provide a {} argument'.format(missing_key),
                'error': 400})

    def test_both_args_kwargs(self):
        body = b64encode(
            create_request_body('the_counter', 1, second_arg=5)
            .encode('ascii'))
        response = self.client.post(self.url,
                                    body,
                                    content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(load_encoded_json(response.content), {})
        self.assertEqual(CALL_COUNTER, 6)

    def test_just_args(self):
        body = b64encode(create_request_body('the_counter', 2).encode('ascii'))
        response = self.client.post(self.url,
                                    body,
                                    content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(load_encoded_json(response.content), {})
        self.assertEqual(CALL_COUNTER, 2)

    def test_just_args_task(self):
        body = b64encode(create_request_body(
            'the_counter_task', 2).encode('ascii'))
        response = self.client.post(self.url,
                                    body,
                                    content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(load_encoded_json(response.content), {})
        self.assertEqual(CALL_COUNTER, 2)

    def test_both_args_kwargs_task(self):
        body = b64encode(
            create_request_body('the_counter_task', 1, second_arg=5)
            .encode('ascii'))
        response = self.client.post(self.url,
                                    body,
                                    content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(load_encoded_json(response.content), {})
        self.assertEqual(CALL_COUNTER, 6)



import json
import boto.sqs

from beanstalk_dispatch.common import create_request_body
from beanstalk_dispatch.execution import execute_function
from django.conf import settings


def schedule_function(queue_name, function_name, *args, **kwargs):
    """
    Schedule a function named `function_name` to be run by workers on
    the queue `queue_name` with *args and **kwargs as specified by that
    function.
    """
    body = create_request_body(function_name, *args, **kwargs)
    if getattr(settings, 'BEANSTALK_DISPATCH_EXECUTE_SYNCHRONOUSLY', False):
        execute_function(json.loads(body))
    else:
        connection = boto.connect_sqs(
            settings.BEANSTALK_DISPATCH_SQS_KEY,
            settings.BEANSTALK_DISPATCH_SQS_SECRET)
        queue = connection.get_queue(queue_name)
        if not queue:
            queue = connection.create_queue(queue_name)
        message = boto.sqs.message.Message()
        message.set_body(body)
        queue.write(message)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-19 20:39
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0040_timeentry_created_at'),
    ]

    operations = [
        migrations.AlterField(
            model_name='payrate',
            name='end_datetime',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='payrate',
            name='start_datetime',
            field=models.DateField(default=django.utils.timezone.now),
        ),
        migrations.RenameField(
            model_name='payrate',
            old_name='end_datetime',
            new_name='end_date',
        ),
        migrations.RenameField(
            model_name='payrate',
            old_name='start_datetime',
            new_name='start_date',
        ),

    ]



from django.core.management.base import BaseCommand
from orchestra.core.errors import WorkflowError
from orchestra.workflow.load import load_workflow


class Command(BaseCommand):
    help = 'Loads a workflow version into the database'

    def add_arguments(self, parser):
        parser.add_argument(
            'app_label',
            help="The app label for the workflow's Django application.")
        parser.add_argument(
            'version_slug',
            help='The version of the workflow to load.')
        parser.add_argument(
            '--force', action='store_true', default=False,
            help=('If a workflow version already exists, update the existing '
                  'version instead of exiting.'))

    def handle(self, *args, **options):
        try:
            load_workflow(options['app_label'],
                          version_slug=options['version_slug'],
                          force=options['force'])
        except WorkflowError as e:
            self.stderr.write(
                'An error occurred while loading the workflow: {}'.format(e))



from django.db import transaction

from orchestra.models import Certification
from orchestra.models import WorkerCertification
from orchestra.models import Workflow

import logging
logger = logging.getLogger(__name__)


@transaction.atomic
def migrate_certifications(source_workflow_slug,
                           destination_workflow_slug,
                           certification_slugs):
    source_worker_certifications = WorkerCertification.objects.filter(
        certification__workflow__slug=source_workflow_slug)

    source_workflow = Workflow.objects.get(slug=source_workflow_slug)
    source_certifications = source_workflow.certifications
    if certification_slugs:
        source_certifications = source_certifications.filter(
            slug__in=certification_slugs)
        not_found_certifications = set(certification_slugs) - set(list(
            source_certifications.values_list('slug', flat=True)))
        if not_found_certifications:
            logger.exception(
                "Certifications {} don't exist for the source workflow".format(
                    not_found_certifications))

    for source_certification in source_certifications.all():
        certification_slug = source_certification.slug
        logger.warning('Migrating {} certifications from {} to {}...'.format(
            certification_slug,
            source_workflow_slug,
            destination_workflow_slug))
        try:
            destination_certification = Certification.objects.get(
                workflow__slug=destination_workflow_slug,
                slug=certification_slug)
        except Certification.DoesNotExist:
            logger.exception(
                ("Certification {} doesn't exist for the destination workflow"
                    .format(certification_slug, destination_workflow_slug)))
            continue

        for worker_certification in source_worker_certifications.filter(
                certification=source_certification).order_by('role'):
            WorkerCertification.objects.get_or_create(
                certification=destination_certification,
                worker=worker_certification.worker,
                task_class=worker_certification.task_class,
                role=worker_certification.role)



from django.core.urlresolvers import reverse
from unittest.mock import patch

from orchestra.tests.helpers import OrchestraAuthenticatedTestCase
from orchestra.models import CommunicationPreference
from orchestra.tests.helpers.fixtures import WorkerFactory
from orchestra.tests.helpers.fixtures import setup_models


class AccountSettingsTest(OrchestraAuthenticatedTestCase):

    def setUp(self):
        super().setUp()
        self.user = self.authenticate_user()
        self.url = reverse('orchestra:account_settings')
        self.worker = WorkerFactory(user=self.user)

    def _get_account_settings_mock_data(self):
        return {
            'first_name': 'Mock first',
            'last_name': 'Mock last',
            'slack_username': 'Mock slack',
            'phone_0': '+1',
            'phone_1': '3477761527',
        }

    def test_get_form(self):
        response = self.request_client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'accounts/settings.html')

    @patch('orchestra.accounts.forms.get_slack_user_id')
    def test_change_all_fields(self, mock_get_slack_user_id):
        mock_get_slack_user_id.return_value = 'test_id'

        data = self._get_account_settings_mock_data()
        response = self.request_client.post(self.url, data)
        self.assertTrue(response.context['success'])

        self.user.refresh_from_db()
        self.assertEqual(self.user.first_name, data['first_name'])
        self.assertEqual(self.user.last_name, data['last_name'])

        self.worker.refresh_from_db()
        self.assertEqual(self.worker.slack_username, data['slack_username'])
        self.assertEqual(self.worker.phone, data['phone_0'] + data['phone_1'])
        self.assertEqual(self.worker.slack_user_id, 'test_id')

    @patch('orchestra.accounts.forms.get_slack_user_id')
    def test_missing_fields(self, mock_get_slack_user_id):
        mock_get_slack_user_id.return_value = 'test_id'

        required_fields = self._get_account_settings_mock_data().keys()
        for field in required_fields:
            data = self._get_account_settings_mock_data()
            data.pop(field)
            response = self.request_client.post(self.url, data)
            self.assertFalse(response.context['success'])


class CommunicationPreferenceSettingsTest(OrchestraAuthenticatedTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.url = reverse('orchestra:communication_preference_settings')
        self.user = self.authenticate_user()

        worker = self.workers[0]
        worker.user = self.user
        worker.save()
        # Update this to handle multiple prefs when we have them
        self.comm_pref = CommunicationPreference.objects.filter(
            worker=worker).first()

    def test_get_form(self):
        response = self.request_client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, 'accounts/communication_preferences_settings.html')

    def _get_mock_data(self):
        return {
            'form-TOTAL_FORMS': 1,
            'form-INITIAL_FORMS': 1,
            'form-MIN_NUM_FORMS': 0,
            'form-MAX_NUM_FORMS': 1,
            'form-0-id': self.comm_pref.id,
        }

    def test_disable_email(self):

        # email is unset and therefore false
        data = self._get_mock_data()
        response = self.request_client.post(self.url, data)
        self.assertTrue(response.context['success'])

        self.comm_pref.refresh_from_db()
        self.assertFalse(self.comm_pref.methods.email.is_set)



from orchestra.models import WorkerCertification
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import CertificationFactory
from orchestra.tests.helpers.fixtures import WorkerFactory
from orchestra.tests.helpers.fixtures import WorkerCertificationFactory
from orchestra.tests.helpers.fixtures import WorkflowFactory
from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.workflow.certifications import migrate_certifications


class ManageCertificationsTestCase(OrchestraTestCase):
    def setUp(self):
        self.workflow_old = WorkflowFactory(
            slug='workflow_old',
            name='Old workflow',
            description='Old workflow to migrate certifications from.',
            code_directory='workflow_old')
        self.workflow_new = WorkflowFactory(
            slug='workflow_new',
            name='New workflow',
            description='New workflow to migrate certifications from.',
            code_directory='workflow_new')

        for workflow in (self.workflow_old, self.workflow_new):
            # Certifications must exist in both workflows for certification
            # to be migrated
            CertificationFactory(
                slug='certification1',
                name='Certification 1',
                description='First certification to migrate.',
                workflow=workflow)
            CertificationFactory(
                slug='certification2',
                name='Certification 2',
                description='Second certification to migrate.',
                workflow=workflow)

        user = (UserFactory(username='test',
                            first_name='test',
                            last_name='test',
                            password='test',
                            email='test@test.com'))
        self.worker = WorkerFactory(user=user)

        for certification in self.workflow_old.certifications.all():
            # Worker certifications exist only for old workflow
            WorkerCertificationFactory(
                worker=self.worker,
                certification=certification,
                role=WorkerCertification.Role.ENTRY_LEVEL)
            WorkerCertificationFactory(
                worker=self.worker,
                certification=certification,
                role=WorkerCertification.Role.REVIEWER)

        super().setUp()

    def test_migrate_certifications(self):
        def _check_old_certifications_unchanged():
            self.assertEqual(WorkerCertification.objects.filter(
                worker=self.worker,
                certification__workflow=self.workflow_old).count(), 4)
            self.assertEqual(WorkerCertification.objects.filter(
                worker=self.worker,
                certification__workflow=self.workflow_old,
                certification__slug='certification1',
                role=WorkerCertification.Role.ENTRY_LEVEL).count(), 1)
            self.assertEqual(WorkerCertification.objects.filter(
                worker=self.worker,
                certification__workflow=self.workflow_old,
                certification__slug='certification1',
                role=WorkerCertification.Role.REVIEWER).count(), 1)
            self.assertEqual(WorkerCertification.objects.filter(
                worker=self.worker,
                certification__workflow=self.workflow_old,
                certification__slug='certification2',
                role=WorkerCertification.Role.ENTRY_LEVEL).count(), 1)
            self.assertEqual(WorkerCertification.objects.filter(
                worker=self.worker,
                certification__workflow=self.workflow_old,
                certification__slug='certification2',
                role=WorkerCertification.Role.REVIEWER).count(), 1)

        _check_old_certifications_unchanged()
        # New workflow should have no worker certifications
        self.assertEqual(WorkerCertification.objects.filter(
            worker=self.worker,
            certification__workflow=self.workflow_new).count(), 0)

        # Migrate `certification1`
        migrate_certifications(
            self.workflow_old.slug,
            self.workflow_new.slug,
            ['certification1'])

        _check_old_certifications_unchanged()

        # New workflow should have only `certification1` worker certifications
        self.assertEqual(WorkerCertification.objects.filter(
            worker=self.worker,
            certification__workflow=self.workflow_new).count(), 2)
        self.assertEqual(WorkerCertification.objects.filter(
            worker=self.worker,
            certification__workflow=self.workflow_new,
            certification__slug='certification1',
            role=WorkerCertification.Role.ENTRY_LEVEL).count(), 1)
        self.assertEqual(WorkerCertification.objects.filter(
            worker=self.worker,
            certification__workflow=self.workflow_new,
            certification__slug='certification1',
            role=WorkerCertification.Role.REVIEWER).count(), 1)

        # Migrate all source certifications
        migrate_certifications(
            self.workflow_old.slug,
            self.workflow_new.slug,
            [])

        _check_old_certifications_unchanged()

        # `certification2` should now be migrated as well
        self.assertEqual(WorkerCertification.objects.filter(
            worker=self.worker,
            certification__workflow=self.workflow_new).count(), 4)
        self.assertEqual(WorkerCertification.objects.filter(
            worker=self.worker,
            certification__workflow=self.workflow_new,
            certification__slug='certification1',
            role=WorkerCertification.Role.ENTRY_LEVEL).count(), 1)
        self.assertEqual(WorkerCertification.objects.filter(
            worker=self.worker,
            certification__workflow=self.workflow_new,
            certification__slug='certification1',
            role=WorkerCertification.Role.REVIEWER).count(), 1)
        self.assertEqual(WorkerCertification.objects.filter(
            worker=self.worker,
            certification__workflow=self.workflow_new,
            certification__slug='certification2',
            role=WorkerCertification.Role.ENTRY_LEVEL).count(), 1)
        self.assertEqual(WorkerCertification.objects.filter(
            worker=self.worker,
            certification__workflow=self.workflow_new,
            certification__slug='certification2',
            role=WorkerCertification.Role.REVIEWER).count(), 1)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2016-11-16 20:32
from __future__ import unicode_literals

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0067_staffbotrequest_last_inquiry_sent'),
    ]

    operations = [
        migrations.AddField(
            model_name='step',
            name='creation_policy',
            field=jsonfield.fields.JSONField(default={}),
        ),
    ]



# -*- coding: utf-8 -*-
# Manually written
from __future__ import unicode_literals

from django.db import migrations, models

import datetime
import dateutil


def create_time_entries(apps, schema_editor):
    TaskAssignment = apps.get_model('orchestra', 'TaskAssignment')
    TimeEntry = apps.get_model('orchestra', 'TimeEntry')

    for assignment in TaskAssignment.objects.all():
        # Check if keys exist before processing, to be compatible with previous
        # versions of snapshots.
        if 'snapshots' not in assignment.snapshots:
            continue
        for snapshot in assignment.snapshots['snapshots']:
            if 'datetime' not in snapshot or 'work_time_seconds' not in snapshot:
                continue
            date = dateutil.parser.parse(snapshot['datetime']).date()
            time_worked = datetime.timedelta(
                seconds=snapshot['work_time_seconds'])
            TimeEntry.objects.get_or_create(assignment=assignment, date=date,
                                            time_worked=time_worked)


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0034_timeentry_is_deleted'),
    ]

    operations = [
        migrations.RunPython(create_time_entries),  # manually-reviewed
    ]



import base64
import json
import os
from collections import defaultdict

from django.conf import settings
from django.contrib.auth import get_user_model
from django.http import HttpResponse
from django.shortcuts import render
from django.contrib.auth.decorators import login_required
from django.templatetags.static import static
from django.views.decorators.csrf import requires_csrf_token
from django_filters import rest_framework as filters
from jsonview.decorators import json_view
from jsonview.exceptions import BadRequest
from rest_framework import generics
from rest_framework import permissions
from rest_framework import status as http_status

from orchestra.core.errors import TaskStatusError
from orchestra.core.errors import TimerError
from orchestra.core.errors import WorkerCertificationError
from orchestra.core.errors import NoTaskAvailable
from orchestra.core.errors import TaskAssignmentError
from orchestra.core.errors import IllegalTaskSubmission
from orchestra.filters import TimeEntryFilter
from orchestra.models import Iteration
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import TimeEntry
from orchestra.models import Worker
from orchestra.models import Step
from orchestra.project_api.serializers import TaskTimerSerializer
from orchestra.project_api.serializers import TimeEntrySerializer
from orchestra.utils.s3 import upload_editor_image
from orchestra.utils import time_tracking
from orchestra.utils.load_json import load_encoded_json
from orchestra.utils.task_lifecycle import save_task
from orchestra.utils.task_lifecycle import submit_task
from orchestra.utils.task_lifecycle import tasks_assigned_to_worker
from orchestra.utils.task_lifecycle import get_new_task_assignment
from orchestra.utils.task_lifecycle import get_task_overview_for_worker
from orchestra.utils.task_lifecycle import worker_assigned_to_rejected_task
from orchestra.utils.task_lifecycle import worker_assigned_to_max_tasks
from orchestra.utils.task_lifecycle import worker_has_reviewer_status
from orchestra.utils.view_helpers import IsAssociatedWorker

import logging
logger = logging.getLogger(__name__)
UserModel = get_user_model()


# NOTE(joshblum): whitenoise is a bit over eager and tries to replace our
# {{js/css}} with a static resource, so we don't put these script tags on the
# page
def _get_script_tag(script):
    return '<script src="{}" type="text/javascript"></script>'.format(script)


def _get_style_tag(style):
    return '<link href="{}" rel="stylesheet">'.format(style)


@login_required
def index(request):
    javascript_includes = []
    stylesheet_includes = []
    orchestra_arguments = {
        'angular_modules': [],
        'angular_directives': defaultdict(lambda: defaultdict(lambda: {})),
        'enable_new_task_buttons': settings.ORCHESTRA_ENABLE_NEW_TASK_BUTTONS,
    }

    for step in Step.objects.filter(is_human=True):
        # Preserve js and stylesheet order while removing duplicates
        for js in step.user_interface.get('javascript_includes', []):
            static_js = _get_script_tag(static(js))
            if static_js not in javascript_includes:
                javascript_includes.append(static_js)
        for style in step.user_interface.get('stylesheet_includes', []):
            static_style = _get_style_tag(static(style))
            if static_style not in stylesheet_includes:
                stylesheet_includes.append(static_style)

        if step.user_interface.get('angular_module'):
            orchestra_arguments['angular_modules'].append(
                step.user_interface['angular_module'])

        if step.user_interface.get('angular_directive'):
            orchestra_arguments['angular_directives'][
                step.workflow_version.workflow.slug][
                    step.workflow_version.slug][step.slug] = (
                        step.user_interface['angular_directive'])

    return render(request, 'orchestra/index.html', {
        'javascript_includes': javascript_includes,
        'stylesheet_includes': stylesheet_includes,
        'orchestra_arguments': json.dumps(orchestra_arguments)})


@json_view
@login_required
def dashboard_tasks(request):
    worker = Worker.objects.get(user=request.user)
    tasks = tasks_assigned_to_worker(worker)
    prevent_new_tasks = (worker_assigned_to_rejected_task(worker) or
                         worker_assigned_to_max_tasks(worker))
    return {'tasks': tasks,
            'preventNew': prevent_new_tasks,
            'reviewerStatus': worker_has_reviewer_status(worker)}


@json_view
@login_required
def new_task_assignment(request, task_type):
    new_tasks_status = {
        'entry_level': Task.Status.AWAITING_PROCESSING,
        'reviewer': Task.Status.PENDING_REVIEW
    }
    try:
        task_status = new_tasks_status[task_type]
    except KeyError:
        raise BadRequest('No such task type')

    worker = Worker.objects.get(user=request.user)
    try:
        task_assignment = get_new_task_assignment(worker, task_status)
    except WorkerCertificationError:
        raise BadRequest('No worker certificates')
    except NoTaskAvailable:
        raise BadRequest('No task')

    task = task_assignment.task
    return {
        'id': task.id,
        'assignment_id': task_assignment.id,
        'step': task.step.slug,
        'project': task.project.workflow_version.slug,
        'detail': task.project.short_description
    }


@json_view
@login_required
def upload_image(request):
    upload_data = load_encoded_json(request.body)
    image_type = upload_data['image_type']
    image_data = base64.b64decode(upload_data['image_data'])
    prefix = upload_data.get('prefix') or ''
    if settings.PRODUCTION:
        prefix = os.path.join('production', prefix)
    else:
        prefix = os.path.join('development', prefix)
    return {'url': upload_editor_image(image_data, image_type, prefix)}


@json_view
@login_required
def task_assignment_information(request):
    try:
        worker = Worker.objects.get(user=request.user)
        return get_task_overview_for_worker(
            load_encoded_json(request.body)['task_id'],
            worker)
    except TaskAssignmentError as e:
        raise BadRequest(e)
    except Task.DoesNotExist as e:
        raise BadRequest(e)


@json_view
@login_required
def save_task_assignment(request):
    assignment_information = load_encoded_json(request.body)
    worker = Worker.objects.get(user=request.user)
    try:
        save_task(assignment_information['task_id'],
                  assignment_information['task_data'],
                  worker)
        return {}
    except Task.DoesNotExist:
        raise BadRequest('No task for given id')
    except TaskAssignmentError as e:
        raise BadRequest(e)


@json_view
@login_required
def submit_task_assignment(request):
    assignment_information = load_encoded_json(request.body)
    worker = Worker.objects.get(user=request.user)
    command_type = assignment_information['command_type']

    if command_type in ('submit', 'accept'):
        iteration_status = Iteration.Status.REQUESTED_REVIEW
    elif command_type == 'reject':
        iteration_status = Iteration.Status.PROVIDED_REVIEW
    else:
        raise BadRequest('Illegal command')

    try:
        submit_task(assignment_information['task_id'],
                    assignment_information['task_data'],
                    iteration_status,
                    worker)
        return {}
    except TaskStatusError:
        raise BadRequest('Task already completed')
    except Task.DoesNotExist:
        raise BadRequest('No task for given id')
    except IllegalTaskSubmission as e:
        raise BadRequest(e)
    except TaskAssignmentError as e:
        raise BadRequest(e)


@json_view
@login_required
def start_timer(request):
    worker = Worker.objects.get(user=request.user)
    try:
        if request.method == 'POST':
            time_entry_data = load_encoded_json(request.body)
            assignment_id = None
            if 'assignment' in time_entry_data:
                assignment_id = time_entry_data['assignment']
            timer = time_tracking.start_timer(worker,
                                              assignment_id=assignment_id)
            serializer = TaskTimerSerializer(timer)
            return serializer.data
    except TaskAssignment.DoesNotExist:
        raise BadRequest('Worker is not assigned to this task id.')
    except TimerError as e:
        raise BadRequest(e)
    except Exception as e:
        logger.exception(e)
        raise e


@json_view
@login_required
def stop_timer(request):
    worker = Worker.objects.get(user=request.user)
    try:
        if request.method == 'POST':
            time_entry = time_tracking.stop_timer(worker)
            serializer = TimeEntrySerializer(time_entry)
            return serializer.data
    except TimerError as e:
        raise BadRequest(e)
    except Exception as e:
        logger.exception(e)
        raise e


@json_view
@login_required
def get_timer(request):
    worker = Worker.objects.get(user=request.user)
    try:
        if request.method == 'GET':
            timer = time_tracking.get_timer_object(worker)
            time_worked = time_tracking.get_timer_current_duration(worker)
            data = TaskTimerSerializer(timer).data
            if time_worked:
                data['time_worked'] = str(time_worked)
            return data
    except Exception as e:
        logger.error(e, exc_info=True)
        raise e


@json_view
@login_required
def update_timer(request):
    worker = Worker.objects.get(user=request.user)
    try:
        if request.method == 'POST':
            data = load_encoded_json(request.body)
            time_tracking.update_timer(
                worker, data.get('description'), data.get('assignment'))
    except Exception as e:
        logger.error(e, exc_info=True)
        raise e


# A simple status endpoint for things like health checks, etc.
def status(request):
    return HttpResponse('OK')


def error_handler(request, error_code, context):
    context.update({
        'contact_us': settings.ORCHESTRA_NOTIFICATIONS_FROM_EMAIL,
        'err_title': error_code,
    })
    return render(request,
                  'orchestra/error.html',
                  context=context,
                  status=error_code)


@requires_csrf_token
def bad_request(request):
    error_code = http_status.HTTP_400_BAD_REQUEST
    return error_handler(request, error_code, context={
        'page_title': '400 Bad Request',
    })


@requires_csrf_token
def forbidden(request):
    error_code = http_status.HTTP_403_FORBIDDEN
    return error_handler(request, error_code, context={
        'page_title': '403 Forbidden',
    })


@requires_csrf_token
def not_found(request):
    error_code = http_status.HTTP_404_NOT_FOUND
    return error_handler(request, error_code, context={
        'page_title': '404 Not Found',
    })


@requires_csrf_token
def internal_server_error(request):
    error_code = http_status.HTTP_500_INTERNAL_SERVER_ERROR
    return error_handler(request, error_code, context={
        'page_title': '500 Internal Server Error',
    })


class TimeEntryList(generics.ListCreateAPIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = TimeEntrySerializer
    filter_backends = (filters.DjangoFilterBackend,)
    filter_class = TimeEntryFilter

    def get_queryset(self):
        """
        Return time entries for current user, filtering on assignment id
        if provided.
        """
        # TODO(lydia): Add time filter to queryset.
        worker = Worker.objects.get(user=self.request.user)
        queryset = TimeEntry.objects.filter(worker=worker)
        assignment_id = self.request.query_params.get('assignment', None)
        if assignment_id is not None:
            queryset = queryset.filter(assignment__id=assignment_id)
        return queryset

    def perform_create(self, serializer):
        """
        Overwrite perform_create so that user can only create time entries
        for him or herself.
        """
        # TODO(lydia): Is there a way to prevent workers from creating
        # time entries for completed TaskAssignments?
        worker = Worker.objects.get(user=self.request.user)
        serializer.save(worker=worker)


class TimeEntryDetail(generics.RetrieveUpdateDestroyAPIView):
    permission_classes = (permissions.IsAuthenticated, IsAssociatedWorker)
    queryset = TimeEntry.objects.all()
    serializer_class = TimeEntrySerializer



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0018_auto_20151007_1451'),
    ]

    operations = [
        migrations.RemoveField(  # manually-reviewed
            model_name='project',
            name='workflow_slug',
        ),
        migrations.RemoveField(  # manually-reviewed
            model_name='task',
            name='step_slug',
        ),
        migrations.AlterField(
            model_name='certification',
            name='workflow',
            field=models.ForeignKey(on_delete=models.CASCADE,
                                    to='orchestra.Workflow', related_name='certifications'),
        ),
        migrations.AlterField(
            model_name='project',
            name='workflow_version',
            field=models.ForeignKey(on_delete=models.CASCADE,
                                    to='orchestra.WorkflowVersion', related_name='projects'),
        ),
        migrations.AlterField(
            model_name='task',
            name='step',
            field=models.ForeignKey(
                on_delete=models.CASCADE, to='orchestra.Step', related_name='tasks'),
        ),
        migrations.AlterField(
            model_name='step',
            name='workflow_version',
            field=models.ForeignKey(on_delete=models.CASCADE,
                                    to='orchestra.WorkflowVersion', related_name='steps'),
        ),
        migrations.AlterField(
            model_name='workflowversion',
            name='workflow',
            field=models.ForeignKey(on_delete=models.CASCADE,
                                    to='orchestra.Workflow', related_name='versions'),
        ),
    ]



# Read permission for any user with the link.
read_with_link_permission = {'withLink': False,
                             'value': 'any',
                             'kind': 'drive#permission',
                             'name': 'Read with link permission',
                             'type': 'anyone',
                             'role': 'reader'}

# Write permission for any user with the link.
write_with_link_permission = {'kind': 'drive#permission',
                              'type': 'anyone',
                              'role': 'writer',
                              'withLink': True}



from contextlib import contextmanager
from datetime import timedelta
from django.conf import settings
from timeout_decorator import timeout as timeout_decorator
from timeout_decorator import TimeoutError

import logging
logger = logging.getLogger(__name__)


@contextmanager
def timeout(function, time):
    yield timeout_decorator(time)(function)


class SafeTask(object):
    """
    Runs a function within a `@timeout` decorator and catches any exceptions.
    timeout_timedelta: maximum number of seconds task can run
    verbose: boolean specifying if failures are logged
    """

    timeout_timedelta = getattr(settings,
                                'BEANSTALK_DISPATCH_TASK_TIMEOUT',
                                timedelta(minutes=2))
    verbose = True

    def run(self, *args, **kwargs):
        """
        Abstract method to fill in with task work
        """
        pass

    def on_error(self, e, *args, **kwargs):
        """
        Run if the task fails for any reason
        """
        pass

    def on_success(self, *args, **kwargs):
        """
        Run after the task completes successfully
        """
        pass

    def on_completion(self, *args, **kwargs):
        """
        Run after each task (after `on_error` or `on_success`)
        """
        pass

    def process(self, *args, **kwargs):
        """
        args: list of arguments for the `runnable`
        kwargs: dictionary of arguments for the `runnable`
        """
        try:
            timeout_seconds = self.timeout_timedelta.total_seconds()
            with timeout(self.run, timeout_seconds) as run:
                run(*args, **kwargs)
        except Exception as e:
            if self.verbose:
                if isinstance(e, TimeoutError):
                    logger.error('SafeTask timed out: %s', e, exc_info=True)
                else:
                    logger.error('Error running SafeTask: %s',
                                 e, exc_info=True)
            self.on_error(e, *args, **kwargs)
        else:
            self.on_success(*args, **kwargs)
        finally:
            self.on_completion(*args, **kwargs)



import inspect

from pydoc import locate
from django.conf import settings

from beanstalk_dispatch import ARGS
from beanstalk_dispatch import FUNCTION
from beanstalk_dispatch import KWARGS
from beanstalk_dispatch import BeanstalkDispatchError
from beanstalk_dispatch.safe_task import SafeTask


def execute_function(function_request):
    """
    Given a request created by
    `beanstalk_dispatch.common.create_request_body`, executes the
    request.  This function is to be run on a beanstalk worker.
    """
    dispatch_table = getattr(settings, 'BEANSTALK_DISPATCH_TABLE', None)

    if dispatch_table is None:
        raise BeanstalkDispatchError('No beanstalk dispatch table configured')
    for key in (FUNCTION, ARGS, KWARGS):
        if key not in function_request.keys():
            raise BeanstalkDispatchError(
                'Please provide a {} argument'.format(key))

    function_path = dispatch_table.get(
        function_request[FUNCTION], ''
    )

    if function_path:
        runnable = locate(function_path)
        if not runnable:
            raise BeanstalkDispatchError(
                'Unable to locate function: {}'.format(function_path))

        args = function_request[ARGS]
        kwargs = function_request[KWARGS]
        if inspect.isclass(runnable):
            if issubclass(runnable, SafeTask):
                task = runnable()
            else:
                raise BeanstalkDispatchError(
                    'Requested task is not a SafeTask subclass: {}'.format(
                        function_request[FUNCTION]))
        else:
            task = SafeTask()
            task.run = runnable
        task.process(*args, **kwargs)
    else:
        raise BeanstalkDispatchError(
            'Requested function not found: {}'.format(
                function_request[FUNCTION]))



import time

from beanstalk_dispatch.safe_task import SafeTask
from datetime import timedelta
from django.test import TestCase
from unittest.mock import MagicMock


def get_mock_task():
    MockSafeTask = SafeTask
    MockSafeTask.run = MagicMock()
    MockSafeTask.on_error = MagicMock()
    MockSafeTask.on_success = MagicMock()
    MockSafeTask.on_completion = MagicMock()
    return MockSafeTask()


def _sleep_too_long():
    time.sleep(10)
    return MagicMock()


class SafeTaskTestCase(TestCase):

    test_error = Exception('Test exception')
    test_args = ['test', 'args']
    test_kwargs = {'test': 'kwargs'}

    """
    Test the basic SafeTask functionality
    """

    def test_task_success(self):
        """
        Test error free task processing
        """
        task = get_mock_task()
        task.process()

        task.run.assert_called_once_with()
        task.on_success.assert_called_once_with()
        task.on_error.assert_not_called()
        task.on_completion.assert_called_once_with()

    def test_task_success_args_kwargs(self):
        task = get_mock_task()
        task.process(*self.test_args, **self.test_kwargs)

        task.run.assert_called_once_with(*self.test_args, **self.test_kwargs)
        task.on_success.assert_called_once_with(
            *self.test_args, **self.test_kwargs)
        task.on_error.assert_not_called()
        task.on_completion.assert_called_once_with(
            *self.test_args, **self.test_kwargs)

    def test_task_error(self):
        """
        Test running a task that fails
        """
        task = get_mock_task()
        task.run.side_effect = self.test_error
        task.process()

        task.run.assert_called_once_with()
        task.on_success.assert_not_called()
        task.on_error.assert_called_once_with(self.test_error)
        task.on_completion.assert_called_once_with()

    def test_task_error_args_kwargs(self):
        task = get_mock_task()
        task.run.side_effect = self.test_error
        task.process(*self.test_args, **self.test_kwargs)

        task.run.assert_called_once_with(*self.test_args, **self.test_kwargs)
        task.on_success.assert_not_called()
        task.on_error.assert_called_once_with(
            self.test_error, *self.test_args, **self.test_kwargs)
        task.on_completion.assert_called_once_with(
            *self.test_args, **self.test_kwargs)

    def test_task_error_timeout(self):
        # Test timeout
        task = get_mock_task()
        task.timeout_timedelta = timedelta(seconds=0.1)
        task.run = MagicMock(side_effect=_sleep_too_long)
        task.process()

        task.run.assert_called_once_with()
        task.on_success.assert_not_called()
        # There is a problem with mocks when asserting the Exception type when
        # it is TimeoutError
        self.assertEqual(task.on_error.call_count, 1)
        self.assertEqual(str(task.on_error.call_args), 'call(TimeoutError())')
        task.on_completion.assert_called_once_with()



from django.conf.urls import url

from orchestra.communication.views import accept_staffing_request_inquiry
from orchestra.communication.views import available_staffing_requests
from orchestra.communication.views import reject_staffing_request_inquiry


urlpatterns = [
    # Interface API
    url(r'^accept_staffing_request_inquiry/(?P<staffing_request_inquiry_id>[0-9]+)/$',  # noqa
        accept_staffing_request_inquiry,
        name='accept_staffing_request_inquiry'),
    url(r'^reject_staffing_request_inquiry/(?P<staffing_request_inquiry_id>[0-9]+)/$',  # noqa
        reject_staffing_request_inquiry,
        name='reject_staffing_request_inquiry'),
    url(r'^available_staffing_requests/',
        available_staffing_requests,
        name='available_staffing_requests'),
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0014_auto_20150717_2015'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='status',
            field=models.IntegerField(
                default=0, choices=[(0, 'Active'), (2, 'Aborted')]),
        ),
        migrations.AlterField(
            model_name='task',
            name='status',
            field=models.IntegerField(choices=[(0, 'Awaiting Processing'), (1, 'Processing'), (2, 'Pending Review'), (
                3, 'Reviewing'), (4, 'Post-review Processing'), (6, 'Aborted'), (5, 'Complete')]),
        ),
        migrations.AlterField(
            model_name='task',
            name='step_slug',
            field=models.CharField(choices=[('media_extraction', 'Media Extraction'), ('design', 'Design'), ('export', 'Export design'), (
                'export_enhancement_step', 'Export enhancement'), ('website_enhancement', 'Enhance')], max_length=200),
        ),
        migrations.AlterField(
            model_name='workercertification',
            name='worker',
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to='orchestra.Worker', related_name='certifications'),
        ),
        migrations.AlterUniqueTogether(  # manually-reviewed
            name='workercertification',
            unique_together=set(
                [('certification', 'worker', 'task_class', 'role')]),
        ),
    ]



from orchestra.models import Iteration
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import TaskTimer
from orchestra.models import TimeEntry
from orchestra.models import WorkerCertification
from rest_framework import serializers


class ProjectSerializer(serializers.ModelSerializer):

    class Meta:
        model = Project
        fields = (
            'id',
            'workflow_slug',
            'workflow_version_slug',
            'short_description',
            'start_datetime',
            'priority',
            'project_data',
            'team_messages_url',
            'task_class',
        )

    workflow_slug = serializers.SerializerMethodField()

    def get_workflow_slug(self, obj):
        return obj.workflow_version.workflow.slug

    workflow_version_slug = serializers.SlugRelatedField(
        source='workflow_version', slug_field='slug', read_only=True)

    task_class = serializers.ChoiceField(
        choices=WorkerCertification.TASK_CLASS_CHOICES)

    project_data = serializers.SerializerMethodField()

    def get_project_data(self, obj):
        """
        This function exists to automatically deserialize the JSON blob from
        the `project_data` JSONField
        """
        return obj.project_data


class TaskSerializer(serializers.ModelSerializer):

    class Meta:
        model = Task
        fields = (
            'id',
            'step_slug',
            'project',
            'status',
            'latest_data',
            'assignments',
            'start_datetime'
        )

    step_slug = serializers.SlugRelatedField(source='step',
                                             slug_field='slug',
                                             read_only=True)

    status = serializers.SerializerMethodField()
    latest_data = serializers.SerializerMethodField()
    assignments = serializers.SerializerMethodField()

    def get_status(self, obj):
        return dict(Task.STATUS_CHOICES).get(obj.status, None)

    def get_latest_data(self, obj):
        """
        Return latest input data for a specified task.

        Args:
            task (orchestra.models.Task):
                The task object for which to retrieve data.

        Returns:
            latest_data (str):
                A serialized JSON blob containing the latest input data.
        """
        active_assignment = (obj.assignments
                             .filter(status=TaskAssignment.Status.PROCESSING))
        if active_assignment.exists():
            assignment = active_assignment[0]
        else:
            assignment = (obj.assignments.all()
                          .order_by('-assignment_counter').first())
        if not assignment:
            return None

        latest_data = assignment.in_progress_task_data
        return latest_data

    def get_assignments(self, obj):
        assignments = TaskAssignmentSerializer(obj.assignments.all()
                                               .order_by('assignment_counter'),
                                               many=True)
        return assignments.data


class TaskAssignmentSerializer(serializers.ModelSerializer):

    class Meta:
        model = TaskAssignment
        fields = (
            'id',
            'start_datetime',
            'worker',
            'task',
            'status',
            'in_progress_task_data',
            'iterations',
        )

    worker = serializers.SerializerMethodField()
    status = serializers.SerializerMethodField()
    in_progress_task_data = serializers.SerializerMethodField()
    iterations = serializers.SerializerMethodField()

    def get_worker(self, obj):
        if not obj.worker:
            return {
                'id': None,
                'username': None,
                'first_name': None,
                'last_name': None,
            }
        return {
            'id': obj.worker.id,
            'username': obj.worker.user.username,
            'first_name': obj.worker.user.first_name,
            'last_name': obj.worker.user.last_name,
        }

    def get_status(self, obj):
        return dict(TaskAssignment.STATUS_CHOICES).get(obj.status, None)

    def get_iterations(self, obj):
        iterations = IterationSerializer(
            obj.iterations.order_by('start_datetime'), many=True)
        return iterations.data

    def get_in_progress_task_data(self, obj):
        """
        This function exists to automatically deserialize the JSON blob from
        the `in_progress_task_data` JSONField

        TODO(derek): maybe make a custom JSON serializer field type?
        """
        return obj.in_progress_task_data


class TimeEntrySerializer(serializers.ModelSerializer):

    class Meta:
        model = TimeEntry
        fields = ('id', 'date', 'time_worked', 'description', 'assignment')
        read_only_fields = ('id',)


class TaskTimerSerializer(serializers.ModelSerializer):

    class Meta:
        model = TaskTimer
        fields = '__all__'


class IterationSerializer(serializers.ModelSerializer):

    class Meta:
        model = Iteration
        fields = (
            'id',
            'start_datetime',
            'end_datetime',
            'status',
            'assignment',
            'submitted_data',
        )

    status = serializers.SerializerMethodField()
    submitted_data = serializers.SerializerMethodField()

    def get_status(self, obj):
        return dict(Iteration.STATUS_CHOICES).get(obj.status, None)

    def get_submitted_data(self, obj):
        """
        This function exists to automatically deserialize the JSON blob from
        the `in_progress_task_data` JSONField
        """
        return obj.submitted_data



import re

from orchestra.bots.errors import SlackCommandInvalidRequest
from orchestra.communication.slack import format_slack_message
from orchestra.communication.slack import OrchestraSlackService


class BaseBot(object):

    # Tuple of (pattern, command_function_string) to match commands to
    commands = ()
    default_error_text = 'Sorry! We couldn\'t process your command'

    def __init__(self, token,
                 allowed_team_ids=None,
                 allowed_domains=None,
                 allowed_channel_ids=None,
                 allowed_channel_names=None,
                 allowed_user_ids=None,
                 allowed_user_names=None,
                 allowed_commands=None,
                 slack_api_key=None,
                 **kwargs):
        """
        Base configuration for the bot to validate messages. If a variable
        is None, all fields are allowed, otherwise each request will only
        be accepted if the request fields are contained in the whitelists.
        """
        self.token = token
        self.slack = OrchestraSlackService(slack_api_key)
        self.whitelists = {
            'team_id': allowed_team_ids,
            'team_domain': allowed_domains,
            'channel_id': allowed_channel_ids,
            'channel_name': allowed_channel_names,
            'user_id': allowed_user_ids,
            'user_name': allowed_user_names,
            'command': allowed_commands,
        }
        self.commands += ((r'help', 'help'),)
        self.command_matchers = [
            (re.compile(pattern, re.IGNORECASE), command)
            for pattern, command in self.commands
        ]

    def validate(self, data):
        """
        Handle a request received from slack. First we validate the
        request and then pass the message to the appropriate handler.
        """
        token = data.get('token')
        if token != self.token:
            raise SlackCommandInvalidRequest('Invalid token.')

        for fieldname, whitelist in self.whitelists.items():
            if whitelist is None:
                continue
            else:
                value = data.get(fieldname)
                if value is None:
                    raise SlackCommandInvalidRequest(
                        '{} is missing'.format(fieldname))

                if value not in whitelist:
                    raise SlackCommandInvalidRequest(
                        'Field {} did not validate: {}'.format(
                            fieldname, value))
        return data

    def help(self):
        """
        Base classes should implement a description of what the command does.
        """
        raise NotImplementedError

    def no_command_found(self, text):
        """
        If we are unable to parse the command, we return this helpful error
        message.
        """
        return format_slack_message(
            '{}: {}'.format(self.default_error_text, text)
        )

    def _find_command(self, text):
        for matcher, command_fn in self.command_matchers:
            match = matcher.match(text)
            if match is not None:
                return getattr(self, command_fn), match.groupdict()

        # If we don't match anything, at least let the user know.
        return self.no_command_found, {'text': text}

    def dispatch(self, data):
        """
        Method to pass data for processing. Should return a dictionary of
        data to return to the user.
        """
        data = self.validate(data)
        text = data.get('text')

        command_fn, kwargs = self._find_command(text)
        return command_fn(**kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def set_task_start_datetime(apps, schema_editor):
    """Fill in legacy `Task.start_datetime` as `Project.start_datetime.`"""
    Task = apps.get_model('orchestra', 'Task')  # noqa
    for task in Task.objects.all():
        task.start_datetime = task.project.start_datetime
        task.save()


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0017_auto_20151012_1719'),
    ]

    operations = [
        migrations.RunPython(set_task_start_datetime),  # manually-reviewed
    ]



from datetime import timedelta
from django.utils import timezone
from unittest.mock import patch

from orchestra.bots.errors import StaffingResponseException
from orchestra.communication.staffing import get_available_requests
from orchestra.communication.staffing import handle_staffing_response
from orchestra.communication.staffing import send_staffing_requests
from orchestra.models import CommunicationPreference
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import StaffingResponse
from orchestra.models import TaskAssignment
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import CommunicationPreferenceFactory
from orchestra.tests.helpers.fixtures import StaffBotRequestFactory
from orchestra.tests.helpers.fixtures import StaffingRequestInquiryFactory
from orchestra.tests.helpers.fixtures import WorkerFactory


class StaffingTestCase(OrchestraTestCase):

    def setUp(self):
        self.worker = WorkerFactory()
        self.staffing_request_inquiry = StaffingRequestInquiryFactory(
            communication_preference__worker=self.worker,
            communication_preference__communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value),
            request__task__step__is_human=True
        )
        super().setUp()

    def test_handle_staffing_response_invalid_request(self):
        old_count = StaffingResponse.objects.all().count()

        # Invalid staffing_request_inquiry_id
        response = handle_staffing_response(
            self.worker, None, is_available=True)
        self.assertEqual(response, None)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count)

        # Invalid worker
        response = handle_staffing_response(
            None, self.staffing_request_inquiry.id, is_available=True)
        self.assertEqual(response, None)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count)

    def test_handle_staffing_response_is_available(self):
        # Test StaffingResponse object creation
        old_count = StaffingResponse.objects.all().count()

        # assign task is called
        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=True)
        self.assertTrue(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.COMPLETE.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        task_assignment = (
            TaskAssignment.objects
            .get(worker=self.worker,
                 task=self.staffing_request_inquiry.request.task))
        self.assertEqual(task_assignment.status,
                         TaskAssignment.Status.PROCESSING)
        self.staffing_request_inquiry.request.task.refresh_from_db()

        # Replay of same request
        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id, is_available=True)
        self.assertTrue(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.COMPLETE.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        # Change mind to `is_available=False` does not do anything
        with self.assertRaises(StaffingResponseException):
            response = handle_staffing_response(
                self.worker, self.staffing_request_inquiry.id,
                is_available=False)

        new_request_inquiry = StaffingRequestInquiryFactory(
            request__task__step__is_human=True
        )
        new_worker = new_request_inquiry.communication_preference.worker
        old_count = StaffingResponse.objects.all().count()
        response = handle_staffing_response(
            new_worker, new_request_inquiry.id, is_available=True)
        self.assertTrue(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.COMPLETE.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        task_assignment = (
            TaskAssignment.objects
            .get(worker=new_worker,
                 task=new_request_inquiry.request.task))
        self.assertEqual(task_assignment.status,
                         TaskAssignment.Status.PROCESSING)

        # restaff
        response.is_winner = False
        response.save()

        worker2 = WorkerFactory()
        staffing_request_inquiry2 = StaffingRequestInquiryFactory(
            communication_preference__worker=worker2,
            request__task=new_request_inquiry.request.task
        )
        response = handle_staffing_response(
            worker2, staffing_request_inquiry2.id, is_available=True)
        self.assertTrue(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.COMPLETE.value)
        task_assignment.refresh_from_db()
        self.assertEqual(task_assignment.worker, worker2)

    def test_handle_staffing_response_not_is_available(self):
        # Test StaffingResponse object creation
        old_count = StaffingResponse.objects.all().count()
        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=False)
        self.assertFalse(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.PROCESSING.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        # Replay of same request
        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=False)
        self.assertFalse(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.PROCESSING.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        # Change mind to `is_available=True`
        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=True)
        self.assertTrue(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.COMPLETE.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

        # Task is not available to claim
        new_request_inquiry = StaffingRequestInquiryFactory(
            request=self.staffing_request_inquiry.request,
            request__task__step__is_human=True)
        new_worker = new_request_inquiry.communication_preference.worker
        old_count = StaffingResponse.objects.all().count()
        response = handle_staffing_response(
            new_worker, new_request_inquiry.id, is_available=True)
        self.assertFalse(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.COMPLETE.value)
        self.assertEqual(StaffingResponse.objects.all().count(), old_count + 1)

    @patch('orchestra.communication.staffing.message_experts_slack_group')
    def test_send_staffing_requests(self, mock_slack):

        worker2 = WorkerFactory()
        CommunicationPreferenceFactory(
            worker=worker2,
            communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value))

        request = StaffBotRequestFactory()
        self.assertEqual(request.status,
                         StaffBotRequest.Status.PROCESSING.value)

        send_staffing_requests(worker_batch_size=1,
                               frequency=timedelta(minutes=0))
        mock_slack.assert_not_called()
        request.refresh_from_db()
        self.assertEqual(request.status,
                         StaffBotRequest.Status.PROCESSING.value)
        # Inquiries increase by two because we send a Slack and an
        # email notification.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            2)

        send_staffing_requests(worker_batch_size=1,
                               frequency=timedelta(minutes=0))
        self.assertEqual(mock_slack.call_count, 1)
        mock_slack.reset()

        request.refresh_from_db()
        self.assertEqual(request.status,
                         StaffBotRequest.Status.PROCESSING.value)
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            4)

        # marked as complete and no new request inquiries sent.
        send_staffing_requests(worker_batch_size=1,
                               frequency=timedelta(minutes=0))
        self.assertTrue(mock_slack.called)
        request.refresh_from_db()
        self.assertEqual(request.status,
                         StaffBotRequest.Status.COMPLETE.value)
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            4)

    @patch('orchestra.communication.staffing.message_experts_slack_group')
    def test_send_staffing_request_priorities(self, mock_slack):

        worker2 = WorkerFactory(staffing_priority=-1)
        CommunicationPreferenceFactory(
            worker=worker2,
            communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value))

        worker3 = WorkerFactory(staffing_priority=1)
        CommunicationPreferenceFactory(
            worker=worker3,
            communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value))

        request = StaffBotRequestFactory()
        self.assertEqual(request.status,
                         StaffBotRequest.Status.PROCESSING.value)

        # Workers should be contacted in priority order.
        excluded = []
        for worker in (worker3, self.worker, worker2):
            send_staffing_requests(worker_batch_size=1,
                                   frequency=timedelta(minutes=0))
            inquiries = (
                StaffingRequestInquiry.objects
                .filter(request=request)
                .exclude(communication_preference__worker__in=excluded))
            for inquiry in inquiries:
                self.assertEqual(worker.id,
                                 inquiry.communication_preference.worker.id)
            excluded.append(worker)

    @patch('orchestra.communication.staffing.message_experts_slack_group')
    def test_handle_staffing_response_all_rejected(self, mock_slack):
        worker2 = WorkerFactory()

        CommunicationPreferenceFactory(
            worker=worker2,
            communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value))

        response = handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=False)
        self.assertFalse(response.is_winner)
        self.assertEqual(response.request_inquiry.request.status,
                         StaffBotRequest.Status.PROCESSING.value)
        self.assertEqual(mock_slack.call_count, 1)
        mock_slack.reset()

        handle_staffing_response(
            worker2, self.staffing_request_inquiry.id,
            is_available=False)

        self.assertEqual(mock_slack.call_count, 1)

    @patch('orchestra.communication.staffing.message_experts_slack_group')
    def test_get_available_request(self, mock_slack):
        # Complete all open requests so new worker doesn't receive them.
        send_staffing_requests(worker_batch_size=2,
                               frequency=timedelta(minutes=0))

        self.assertEqual(len(get_available_requests(self.worker)), 1)

        worker2 = WorkerFactory()
        CommunicationPreferenceFactory(
            worker=worker2,
            communication_type=(
                CommunicationPreference.CommunicationType
                .NEW_TASK_AVAILABLE.value))
        request1 = StaffBotRequestFactory(task__step__is_human=True)
        request2 = StaffBotRequestFactory(task__step__is_human=True)

        send_staffing_requests(worker_batch_size=2,
                               frequency=timedelta(minutes=0))
        inquiry1 = (
            StaffingRequestInquiry.objects
            .filter(communication_preference__worker=self.worker)
            .filter(request=request1).first())
        inquiry2 = (
            StaffingRequestInquiry.objects
            .filter(communication_preference__worker=worker2)
            .filter(request=request2).first())

        # `self.worker` now has three available tasks, whereas `worker2`
        # just has access to the two new tasks.
        available_requests = get_available_requests(self.worker)
        self.assertEqual(len(available_requests), 3)
        self.assertEqual(len(get_available_requests(worker2)), 2)

        # Tasks should be sorted by start_datetime in ascending order.

        first_task, second_task, third_task = (
            available_requests[0]['task'], available_requests[1]['task'],
            available_requests[2]['task'])
        self.assertLess(
            first_task.start_datetime, second_task.start_datetime)
        self.assertLess(
            second_task.start_datetime, third_task.start_datetime)

        # `self.worker` will lose an available task (they accept it),
        # whereas `worker2` is unchanged.
        handle_staffing_response(
            self.worker, self.staffing_request_inquiry.id,
            is_available=True)
        self.assertEqual(len(get_available_requests(self.worker)), 2)
        self.assertEqual(len(get_available_requests(worker2)), 2)

        # `self.worker` will lose an available task (they ignore it),
        # whereas `worker2` is unchanged.
        handle_staffing_response(
            self.worker, inquiry1.id,
            is_available=False)
        self.assertEqual(len(get_available_requests(self.worker)), 1)
        self.assertEqual(len(get_available_requests(worker2)), 2)

        # `worker2` takes a task.
        handle_staffing_response(
            worker2, inquiry2.id,
            is_available=True)
        self.assertEqual(len(get_available_requests(self.worker)), 0)
        self.assertEqual(len(get_available_requests(worker2)), 1)

    @patch('orchestra.communication.staffing.message_experts_slack_group')
    def test_send_staffing_requests_parameters(self, mock_slack):
        for idx in range(5):
            CommunicationPreferenceFactory(
                worker=WorkerFactory(),
                communication_type=(
                    CommunicationPreference.CommunicationType
                    .NEW_TASK_AVAILABLE.value))
        # Make a new request and turn off the global request, as the
        # global one has an inquiry on it already.
        request = StaffBotRequestFactory(task__step__is_human=True)
        self.staffing_request_inquiry.request.status = (
            StaffBotRequest.Status.COMPLETE)

        send_staffing_requests(worker_batch_size=1,
                               frequency=timedelta(minutes=20))
        # Inquiries increase by two because we send a Slack and an
        # email notification.  `last_inquiry_sent` is None on new
        # tasks, so we send this batch regardless of frequency.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            2)

        send_staffing_requests(worker_batch_size=2,
                               frequency=timedelta(minutes=20))
        # Don't send more inquiries, since it hasn't been 20 minutes.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            2)

        request.last_inquiry_sent = timezone.now() - timedelta(minutes=21)
        request.save()
        send_staffing_requests(worker_batch_size=2,
                               frequency=timedelta(minutes=20))
        # Send two email and two slack inquiries since it's been 21 minutes.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            6)

        send_staffing_requests(worker_batch_size=10,
                               frequency=timedelta(minutes=20))
        # Don't send more inquiries, since it hasn't been 20 minutes.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            6)

        request.last_inquiry_sent = timezone.now() - timedelta(minutes=21)
        request.save()
        send_staffing_requests(worker_batch_size=10,
                               frequency=timedelta(minutes=20))
        # Send remaining inquiries, since enough time has passed.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            12)

        request.last_inquiry_sent = timezone.now() - timedelta(minutes=21)
        request.save()
        send_staffing_requests(worker_batch_size=10,
                               frequency=timedelta(minutes=20))
        # We're all out of workers to whom we'd like to send inquiries.
        self.assertEqual(
            StaffingRequestInquiry.objects.filter(request=request).count(),
            12)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-28 17:50
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0048_merge'),
    ]

    operations = [
        migrations.AddField(
            model_name='staffingrequest',
            name='communication_method',
            field=models.IntegerField(
                choices=[(0, 'slack'), (1, 'email')], default=0),
        ),
        migrations.AddField(
            model_name='staffingrequest',
            name='status',
            field=models.IntegerField(
                choices=[(0, 'pending'), (1, 'sent')], default=0),
        ),
    ]



from apiclient import errors
from apiclient.http import MediaFileUpload
from apiclient.discovery import build
from httplib2 import Http
from oauth2client.service_account import ServiceAccountCredentials

import logging
import re


logger = logging.getLogger(__name__)
_image_mimetype_regex = re.compile('(image/(?:jpg|jpeg|gif|png))',
                                   re.IGNORECASE)


class Service(object):

    def __init__(self, google_p12_path, google_service_email):
        self._service = self._create_drive_service(google_p12_path,
                                                   google_service_email)

    def _create_drive_service(self, google_p12_path,
                              google_service_email):
        credentials = ServiceAccountCredentials.from_p12_keyfile(
            google_service_email,
            google_p12_path,
            scopes=['https://www.googleapis.com/auth/drive'])
        http_auth = credentials.authorize(Http())
        return build('drive', 'v2', http=http_auth, cache_discovery=False)

    def insert_file(self, title, description, parent_id, mime_type, filename):
        """Insert new file.

        Args:
          service: Drive API service instance.
          title: Title of the file to insert, including the extension.
          description: Description of the file to insert.
          parent_id: Parent folder's ID.
          mime_type: MIME type of the file to insert.
          filename: Filename of the file to insert.
        Returns:
          Inserted file metadata if successful, None otherwise.
        """
        media_body = MediaFileUpload(filename, mimetype=mime_type,
                                     resumable=True)

        body = {
            'title': title,
            'description': description,
            'mimeType': mime_type
        }

        # Set the parent folder.
        if parent_id:
            body['parents'] = [{'id': parent_id}]

        try:
            file = self._service.files().insert(
                body=body,
                media_body=media_body).execute()
            return file
        except errors.HttpError:
            logger.exception('An error while creating a file.')
            return None

    def copy_file(self, origin_file_id, copy_title, parent_ids=None):
        """
        Copy an existing file.
        """
        body = {'title': copy_title}
        if parent_ids is not None:
            body.update({'parents': [{'id': parent} for parent in parent_ids]})
        try:
            return self._service.files().copy(fileId=origin_file_id,
                                              body=body).execute()
        except errors.HttpError:
            logger.exception('An error occurred while copying file')
            return None

    def insert_folder(self, title, parent_id):
        """Insert new folder.

        Args:
          service: Drive API service instance.
          title: Title of the folder to insert, including the extension.
          parent_id: Parent folder id
          permissions: List of permissions
        Returns:
          Inserted file metadata if successful, None otherwise.
        """
        body = {
            'title': title,
            'mimeType': 'application/vnd.google-apps.folder'
        }

        # Set the parent folder.
        if parent_id:
            body['parents'] = [{'id': parent_id}]

        try:
            folder = self._service.files().insert(
                body=body).execute()
            return folder
        except errors.HttpError:
            logger.exception('An error while creating a folder.')
            return None

    def add_permission(self, file_id, permission):
        """
        Add permission of a certain type to a given file.
        """
        try:
            return (self._service.permissions()
                    .insert(fileId=file_id, body=permission)
                    .execute())
        except errors.HttpError as error:
            logger.exception('An error occured: %s', error)
            return None

    def delete_folder(self, folder_id):
        """
        Delete folder from a google drive.
        """
        try:
            return (self._service.files()
                    .delete(fileId=folder_id).execute())
        except errors.HttpError as error:
            logger.exception('An error occured: %s', error)
            return None

    def list_folder(self, folder_id):
        """Retrieve a list of File resources.

        Args:
            folder_id: ID of parent folder to list.
        Returns:
            List of File resources.
        """
        service = self._service
        result = []
        page_token = None
        while True:
            try:
                param = {
                    'q': "'{}' in parents".format(folder_id),
                }
                if page_token:
                    param['pageToken'] = page_token
                files = service.files().list(**param).execute()

                result.extend(files['items'])
                page_token = files.get('nextPageToken')
                if not page_token:
                    break
            except errors.HttpError as error:
                logger.exception('An error occurred: %s', error)
                return None
        return result

    def get_file_content(self, file_id):
        """Get a file's content.

        Args:
            file_id: ID of the file to get.

        Returns:
            File's content if successful, None otherwise.
        """
        try:
            return self._service.files().get_media(fileId=file_id).execute()
        except errors.HttpError as error:
            logger.exception('An error occurred: %s', error)
            return None
